from bayserver_core.util.char_util import CharUtil

from bayserver_docker_fcgi.fcg_command import FcgCommand
from bayserver_docker_fcgi.fcg_type import FcgType


#
#  FCGI spec
#    http://www.mit.edu/~yandros/doc/specs/fcgi-spec.html
#
#  Endrequest command format
#          typedef struct {
#              unsigned char appStatusB3;
#              unsigned char appStatusB2;
#              unsigned char appStatusB1;
#              unsigned char appStatusB0;
#              unsigned char protocolStatus;
#              unsigned char reserved[3];
#          } FCGI_EndRequestBody;
#

class CmdEndRequest(FcgCommand):
    FCGI_REQUEST_COMPLETE = 0
    FCGI_CANT_MPX_CONN = 1
    FCGI_OVERLOADED = 2
    FCGI_UNKNOWN_ROLE = 3

    RESERVED = [CharUtil.SPACE_BYTE] * 3

    def __init__(self, req_id):
        super().__init__(FcgType.END_REQUEST, req_id)
        self.app_status = 0
        self.protocol_status = CmdEndRequest.FCGI_REQUEST_COMPLETE

    def unpack(self, pkt):
        super().unpack(pkt)
        acc = pkt.new_data_accessor()
        self.app_status = acc.get_int()
        self.protocol_status = acc.get_byte()

    def pack(self, pkt):
        acc = pkt.new_data_accessor()
        acc.put_int(self.app_status)
        acc.put_byte(self.protocol_status)
        acc.put_bytes(CmdEndRequest.RESERVED)

        # must be called from last line
        super().pack(pkt)

    def handle(self, cmd_handler):
        return cmd_handler.handle_end_request(self)



from bayserver_docker_fcgi.fcg_command import FcgCommand


class InOutCommandBase(FcgCommand):

    def __init__(self, typ, req_id, data=None, start=0, length=0):
        super().__init__(typ, req_id)
        self.data = data
        self.start = start
        self.length = length

    def unpack(self, pkt):
        super().unpack(pkt)
        self.start = pkt.header_len
        self.length = pkt.data_len()
        self.data = pkt.buf

    def pack(self, pkt):
        if self.data is not None and self.length > 0:
            acc = pkt.new_data_accessor()
            acc.put_bytes(self.data, self.start, self.length)

        # must be called from last line
        super().pack(pkt)



from bayserver_core.protocol.packet_store import PacketStore
from bayserver_core.protocol.protocol_handler_store import ProtocolHandlerStore

from bayserver_core.docker.base.port_base import PortBase

from bayserver_docker_fcgi.fcg_docker import FcgDocker
from bayserver_docker_fcgi.fcg_packet_factory import FcgPacketFactory
from bayserver_docker_fcgi.fcg_inbount_handler import FcgInboundHandler


class FcgPortDocker(PortBase, FcgDocker):

    ######################################################
    # Implements Port
    ######################################################
    def protocol(self):
        return super().PROTO_NAME

    ######################################################
    # Implements PortBase
    ######################################################
    def support_anchored(self):
        return True

    def support_unanchored(self):
        return False

    ######################################################
    # Class initializer
    ######################################################
    PacketStore.register_protocol(
        FcgDocker.PROTO_NAME,
        FcgPacketFactory())
    ProtocolHandlerStore.register_protocol(
        FcgDocker.PROTO_NAME,
        True,
        FcgInboundHandler.InboundProtocolHandlerFactory())


from bayserver_core.protocol.protocol_handler import ProtocolHandler
from bayserver_core.protocol.packet_packer import PacketPacker
from bayserver_core.protocol.command_packer import CommandPacker


from bayserver_docker_fcgi.fcg_docker import FcgDocker
from bayserver_docker_fcgi.fcg_packet import FcgPacket
from bayserver_docker_fcgi.fcg_command_unpacker import FcgCommandUnPacker
from bayserver_docker_fcgi.fcg_packet_unpacker import FcgPacketUnPacker


class FcgProtocolHandler(ProtocolHandler, FcgCommandUnPacker):

    def __init__(self, pkt_store, svr_mode):
        super().__init__()
        self.command_unpacker = FcgCommandUnPacker(self)
        self.packet_unpacker = FcgPacketUnPacker(pkt_store, self.command_unpacker)
        self.packet_packer = PacketPacker()
        self.command_packer = CommandPacker(self.packet_packer, pkt_store)
        self.server_mode = svr_mode


    def __str__(self):
        return f"PH[{self.ship}]"

    ######################################################
    # Implements ProtocolHandler
    ######################################################
    def protocol(self):
        return FcgDocker.PROTO_NAME

    def max_req_packet_data_size(self):
        return FcgPacket.MAXLEN

    def max_res_packet_data_size(self):
        return FcgPacket.MAXLEN





from bayserver_core.bay_log import BayLog
from bayserver_core.util.string_util import StringUtil

from bayserver_docker_fcgi.fcg_command import FcgCommand
from bayserver_docker_fcgi.fcg_type import FcgType


#
#  FCGI spec
#    http://www.mit.edu/~yandros/doc/specs/fcgi-spec.html
#
#
#  Params command format (Name-Value list)
#
#          typedef struct {
#              unsigned char nameLengthB0;  // nameLengthB0  >> 7 == 0
#              unsigned char valueLengthB0; // valueLengthB0 >> 7 == 0
#              unsigned char nameData[nameLength];
#              unsigned char valueData[valueLength];
#          } FCGI_NameValuePair11;
#
#          typedef struct {
#              unsigned char nameLengthB0;  // nameLengthB0  >> 7 == 0
#              unsigned char valueLengthB3; // valueLengthB3 >> 7 == 1
#              unsigned char valueLengthB2;
#              unsigned char valueLengthB1;
#              unsigned char valueLengthB0;
#              unsigned char nameData[nameLength];
#              unsigned char valueData[valueLength
#                      ((B3 & 0x7f) << 24) + (B2 << 16) + (B1 << 8) + B0];
#          } FCGI_NameValuePair14;
#
#          typedef struct {
#              unsigned char nameLengthB3;  // nameLengthB3  >> 7 == 1
#              unsigned char nameLengthB2;
#              unsigned char nameLengthB1;
#              unsigned char nameLengthB0;
#              unsigned char valueLengthB0; // valueLengthB0 >> 7 == 0
#              unsigned char nameData[nameLength
#                      ((B3 & 0x7f) << 24) + (B2 << 16) + (B1 << 8) + B0];
#              unsigned char valueData[valueLength];
#          } FCGI_NameValuePair41;
#
#          typedef struct {
#              unsigned char nameLengthB3;  // nameLengthB3  >> 7 == 1
#              unsigned char nameLengthB2;
#              unsigned char nameLengthB1;
#              unsigned char nameLengthB0;
#              unsigned char valueLengthB3; // valueLengthB3 >> 7 == 1
#              unsigned char valueLengthB2;
#              unsigned char valueLengthB1;
#              unsigned char valueLengthB0;
#              unsigned char nameData[nameLength
#                      ((B3 & 0x7f) << 24) + (B2 << 16) + (B1 << 8) + B0];
#              unsigned char valueData[valueLength
#                      ((B3 & 0x7f) << 24) + (B2 << 16) + (B1 << 8) + B0];
#          } FCGI_NameValuePair44;
#

class CmdParams(FcgCommand):

    def __init__(self, req_id):
        super().__init__(FcgType.PARAMS, req_id)
        self.params = []

    def __str__(self):
        return f"FcgCmdParams{self.params}"

    def unpack(self, pkt):
        super().unpack(pkt)
        acc = pkt.new_data_accessor()
        while acc.pos < pkt.data_len():
            name_len = self.read_length(acc)
            value_len = self.read_length(acc)

            name = bytearray(name_len)
            acc.get_bytes(name, 0, name_len)

            value = bytearray(value_len)
            acc.get_bytes(value, 0, value_len)

            name = StringUtil.from_bytes(name)
            value = StringUtil.from_bytes(value)
            BayLog.trace("Params: %s=%s", name, value)
            self.add_param(name, value)

    def pack(self, pkt):
        acc = pkt.new_data_accessor()
        for nv in self.params:
            name = nv[0]
            value = nv[1]
            name_len = len(name)
            value_len = len(value)

            self.write_length(name_len, acc)
            self.write_length(value_len, acc)

            acc.put_string(name)
            acc.put_string(value)

        # must be called from last line
        super().pack(pkt)

    def handle(self, cmd_handler):
        return cmd_handler.handle_params(self)


    #
    # private
    #
    def read_length(self, acc):
        len = acc.get_byte()
        if len >> 7 == 0:
            return len
        else:
            len2 = acc.get_byte()
            len3 = acc.get_byte()
            len4 = acc.get_byte()
            return ((len & 0x7f) << 24) | (len2 << 16) | (len3 << 8) | len4

    def write_length(self, length, acc):
        if length >> 7 == 0:
            acc.put_byte(length)
        else:
            buf = bytearray(4)
            buf[0] = (length >> 24 & 0xFF) | 0x80
            buf[1] = length >> 16 & 0xFF
            buf[2] = length >> 8 & 0xFF
            buf[3] = length & 0xFF
            acc.put_bytes(buf)

    def add_param(self, name, value):
        if name is None:
            raise RuntimeError("nil argument")

        if value is None:
            value = ""

        self.params.append([name, value])







from bayserver_core.protocol.command_unpacker import CommandUnPacker

from bayserver_docker_fcgi.fcg_type import FcgType
from bayserver_docker_fcgi.command.cmd_begin_request import CmdBeginRequest
from bayserver_docker_fcgi.command.cmd_end_request import CmdEndRequest
from bayserver_docker_fcgi.command.cmd_params import CmdParams
from bayserver_docker_fcgi.command.cmd_stdin import CmdStdIn
from bayserver_docker_fcgi.command.cmd_stdout import CmdStdOut
from bayserver_docker_fcgi.command.cmd_stderr import CmdStdErr

class FcgCommandUnPacker(CommandUnPacker):

    def __init__(self, handler):
        self.handler = handler
        self.reset()

    def reset(self):
        pass

    def packet_received(self, pkt):

        if pkt.type == FcgType.BEGIN_REQUEST:
            cmd = CmdBeginRequest(pkt.req_id)

        elif pkt.type == FcgType.END_REQUEST:
            cmd = CmdEndRequest(pkt.req_id)

        elif pkt.type == FcgType.PARAMS:
            cmd = CmdParams(pkt.req_id)

        elif pkt.type == FcgType.STDIN:
            cmd = CmdStdIn(pkt.req_id)

        elif pkt.type == FcgType.STDOUT:
            cmd = CmdStdOut(pkt.req_id)

        elif pkt.type == FcgType.STDERR:
            cmd = CmdStdErr(pkt.req_id)

        else:
            raise RuntimeError("IllegalState")

        cmd.unpack(pkt)
        cmd.handle(self.handler)


from abc import ABCMeta, abstractmethod

from bayserver_core.protocol.command_handler import CommandHandler

class FcgCommandHandler(CommandHandler, metaclass=ABCMeta):

    @abstractmethod
    def handle_begin_request(self, cmd):
        pass

    @abstractmethod
    def handle_end_request(self, cmd):
        pass

    @abstractmethod
    def handle_params(self, cmd):
        pass

    @abstractmethod
    def handle_stderr(self, cmd):
        pass

    @abstractmethod
    def handle_stdin(self, cmd):
        pass

    @abstractmethod
    def handle_stdout(self, cmd):
        pass




from bayserver_docker_fcgi.fcg_type import FcgType
from bayserver_docker_fcgi.command.in_out_command_base import InOutCommandBase

#
#  FCGI spec
#    http://www.mit.edu/~yandros/doc/specs/fcgi-spec.html
#
#   StdErr command format
#    raw data
#

class CmdStdErr(InOutCommandBase):

    def __init__(self, req_id):
        super().__init__(FcgType.STDERR, req_id)

    def handle(self, cmd_handler):
        return cmd_handler.handle_stderr(self)




class FcgType:
    BEGIN_REQUEST = 1
    ABORT_REQUEST = 2
    END_REQUEST = 3
    PARAMS = 4
    STDIN = 5
    STDOUT = 6
    STDERR = 7
    DATA = 8
    GET_VALUES = 9
    GET_VALUES_RESULT = 10
    UNKNOWN_TYPE = 11


import datetime

from bayserver_core.bayserver import BayServer
from bayserver_core.bay_log import BayLog
from bayserver_core.sink import Sink

from bayserver_core.agent.next_socket_action import NextSocketAction
from bayserver_core.protocol.protocol_handler_factory import ProtocolHandlerFactory
from bayserver_core.protocol.protocol_exception import ProtocolException
from bayserver_core.tour.tour import Tour
from bayserver_core.docker.warp.warp_data import WarpData

from bayserver_core.util.char_util import CharUtil
from bayserver_core.util.headers import Headers
from bayserver_core.util.string_util import StringUtil
from bayserver_core.util.cgi_util import CgiUtil

from bayserver_docker_fcgi.fcg_protocol_handler import FcgProtocolHandler
from bayserver_docker_fcgi.command.cmd_stdin import CmdStdIn
from bayserver_docker_fcgi.command.cmd_begin_request import CmdBeginRequest
from bayserver_docker_fcgi.command.cmd_params import CmdParams
from bayserver_docker_fcgi.fcg_params import FcgParams

class FcgWarpHandler(FcgProtocolHandler):
    class WarpProtocolHandlerFactory(ProtocolHandlerFactory):

        def create_protocol_handler(self, pkt_store):
            return FcgWarpHandler(pkt_store)

    STATE_READ_HEADER = 1
    STATE_READ_CONTENT = 2

    def __init__(self, pkt_store):
        super().__init__(pkt_store, False)
        self.cur_warp_id = 0

        self.state = None
        self.line_buf = bytearray()
        self.pos = None
        self.last = None
        self.data = None
        self.reset()

    def reset(self):
        super().reset()
        self.reset_state()
        self.line_buf.clear()
        self.pos = 0
        self.last = 0
        self.data = None


    ######################################################
    # Implements WarpHandler
    ######################################################
    def next_warp_id(self):
        self.cur_warp_id += 1
        return self.cur_warp_id

    def new_warp_data(self, warp_id):
        return WarpData(self.ship, warp_id)

    def post_warp_headers(self, tur):
        self.send_begin_req(tur)
        self.send_params(tur)

    def post_warp_contents(self, tur, buf, start, length, callback):
        self.send_stdin(tur, buf, start, length, callback)

    def post_warp_end(self, tur):
        def callback():
            self.ship.agent.non_blocking_handler.ask_to_read(self.ship.socket)

        self.send_stdin(tur, None, 0, 0, callback)

    def verify_protocol(proto):
        pass

    ######################################################
    # Implements FcgCommandHandler
    ######################################################

    def handle_begin_request(self, cmd):
        raise ProtocolException("Invalid FCGI command: %d", cmd.type)

    def handle_end_request(self, cmd):
        tur = self.ship.get_tour(cmd.req_id)
        self.end_req_content(tur)
        return NextSocketAction.CONTINUE

    def handle_params(self, cmd):
        raise ProtocolException("Invalid FCGI command: %d", cmd.type)

    def handle_stderr(self, cmd):
        msg = cmd.data[cmd.start: cmd.start + cmd.length + 1]
        BayLog.error("%s server error: %s", self, msg)
        return NextSocketAction.CONTINUE

    def handle_stdin(self, cmd):
        raise ProtocolException("Invalid FCGI command: %d", cmd.type)

    def handle_stdout(self, cmd):
        BayLog.debug("%s handle_stdout req_id=%d len=%d", self.ship, cmd.req_id, cmd.length)

        tur = self.ship.get_tour(cmd.req_id)
        if tur is None:
            raise Sink("Tour not found")

        if cmd.length == 0:
            # stdout end
            self.reset_state()
            return NextSocketAction.CONTINUE

        self.data = cmd.data
        self.pos = cmd.start
        self.last = cmd.start + cmd.length

        if self.state == FcgWarpHandler.STATE_READ_HEADER:
            self.read_header(tur)

        if self.pos < self.last:
            BayLog.debug("%s fcgi: pos=%d last=%d len=%d", self.ship, self.pos, self.last, self.last - self.pos)
            if self.state == FcgWarpHandler.STATE_READ_CONTENT:
                available = tur.res.send_content(Tour.TOUR_ID_NOCHECK, self.data, self.pos, self.last - self.pos)
                if not available:
                    return NextSocketAction.SUSPEND

        return NextSocketAction.CONTINUE

    ######################################################
    # Custom methods
    ######################################################
    def read_header(self, tur):
        wdat = WarpData.get(tur)

        header_finished = self.parse_header(wdat.res_headers)
        if header_finished:
            wdat.res_headers.copy_to(tur.res.headers)

            # Check HTTP Status from headers
            status = wdat.res_headers.get(Headers.STATUS)
            if StringUtil.is_set(status):
                stlist = status.split(" ")
                tur.res.headers.status = int(stlist[0])
                tur.res.headers.remove(Headers.STATUS)


            BayLog.debug("%s fcgi: read header status=%d contlen=%d", self.ship, tur.res.headers.status, wdat.res_headers.content_length())
            sid = self.ship.ship_id

            def callback(length, resume):
                if resume:
                    self.ship.resume(sid)

            tur.res.set_consume_listener(callback)

            tur.res.send_headers(Tour.TOUR_ID_NOCHECK)
            self.change_state(FcgWarpHandler.STATE_READ_CONTENT)

    def read_content(self, tur):
        tur.res.send_content(Tour.TOUR_ID_NOCHECK, self.data, self.pos, self.last - self.pos)

    def parse_header(self, headers):

        while True:
            if self.pos == self.last:
                # no byte data
                break

            c = self.data[self.pos]
            self.pos += 1

            if c == CharUtil.CR_BYTE:
                continue
            elif c == CharUtil.LF_BYTE:
                line = StringUtil.from_bytes(self.line_buf)

                if len(line) == 0:
                    return True

                colon_pos = line.find(':')
                if colon_pos == -1:
                    raise ProtocolException("fcgi: Header line of server is invalid: %s", line)
                else:
                    name = line[0:colon_pos].strip()
                    value = line[colon_pos + 1:].strip()

                    if StringUtil.is_empty(name) or StringUtil.is_empty(value):
                        raise ProtocolException("fcgi: Header line of server is invalid: %s", line)

                    headers.add(name, value)
                    if BayServer.harbor.trace_header:
                        BayLog.info("%s fcgi_warp: resHeader: %s=%s", self.ship, name, value)

                self.line_buf.clear()

            else:
                self.line_buf.append(c)

        return False

    def end_req_content(self, tur):
        self.ship.end_warp_tour(tur)
        tur.res.end_content(Tour.TOUR_ID_NOCHECK)
        self.reset_state()

    def change_state(self, new_state):
        self.state = new_state

    def reset_state(self):
        self.change_state(self.STATE_READ_HEADER)

    def send_stdin(self, tur, data, ofs, length, callback):
        cmd = CmdStdIn(WarpData.get(tur).warp_id, data, ofs, length)
        self.ship.post(cmd, callback)

    def send_begin_req(self, tur):
        cmd = CmdBeginRequest(WarpData.get(tur).warp_id)
        cmd.role = CmdBeginRequest.FCGI_RESPONDER
        cmd.keep_conn = True
        self.ship.post(cmd)

    def send_params(self, tur):
        script_base = self.ship.docker.script_base
        if script_base is None:
            script_base = tur.town.location

        if StringUtil.is_empty(script_base):
            raise RuntimeError(f"{tur.town} Could not create SCRIPT_FILENAME. Location of town not specified.")

        doc_root = self.ship.docker.doc_root
        if doc_root is None:
            doc_root = tur.town.location

        if StringUtil.is_empty(doc_root):
            raise RuntimeError(f"{tur.town} docRoot of fcgi docker or location of town is not specified.")

        warp_id = WarpData.get(tur).warp_id
        cmd = CmdParams(warp_id)
        script_fname = [None]

        def callback(name, value):
            if name == CgiUtil.SCRIPT_FILENAME:
                script_fname[0] = value
            else:
                cmd.add_param(name, value)

        CgiUtil.get_env(tur.town.name, doc_root, script_base, tur, callback)

        script_fname = f"proxy:fcgi://{self.ship.docker.host}:{self.ship.docker.port}{script_fname[0]}"
        cmd.add_param(CgiUtil.SCRIPT_FILENAME, script_fname[0])

        # Add FCGI params
        cmd.add_param(FcgParams.CONTEXT_PREFIX, "")
        cmd.add_param(FcgParams.UNIQUE_ID, str(datetime.time()))
        # cmd.add_param(FcgParams.X_FORWARDED_FOR, tur.req.remote_address)
        # cmd.add_param(FcgParams.X_FORWARDED_PROTO, tur.is_secure ? "https" : "http")
        # cmd.add_param(FcgParams.X_FORWARDED_PORT, tur.req.req_port.to_s)

        if BayServer.harbor.trace_header:
            for kv in cmd.params:
                BayLog.info("%s fcgi_warp: env: %s=%s", self.ship, kv[0], kv[1])

        self.ship.post(cmd)

        cmd_params_end = CmdParams(WarpData.get(tur).warp_id)
        self.ship.post(cmd_params_end)







from bayserver_core.protocol.packet_factory import PacketFactory

from bayserver_docker_fcgi.fcg_packet import FcgPacket

class FcgPacketFactory(PacketFactory):

    def create_packet(self, type):
        return FcgPacket(type)


from bayserver_core.protocol.packet import Packet

#
#  FCGI spec
#    http://www.mit.edu/~yandros/doc/specs/fcgi-spec.html
#
#  FCGI Packet (Record) format
#          typedef struct {
#              unsigned char version;
#              unsigned char type;
#              unsigned char requestIdB1;
#              unsigned char requestIdB0;
#              unsigned char contentLengthB1;
#              unsigned char contentLengthB0;
#              unsigned char paddingLength;
#              unsigned char reserved;
#              unsigned char contentData[contentLength];
#              unsigned char paddingData[paddingLength];
#          } FCGI_Record;
#

class FcgPacket(Packet):
    PREAMBLE_SIZE = 8

    VERSION = 1
    MAXLEN = 65535

    FCGI_NULL_REQUEST_ID = 0

    def __init__(self, type):
        super().__init__(type, FcgPacket.PREAMBLE_SIZE, FcgPacket.MAXLEN)
        self.version = FcgPacket.VERSION
        self.req_id = None

    def reset(self):
        super().reset()
        self.version = FcgPacket.VERSION
        self.req_id = 0

    def __str__(self):
        return f"FcgPacket({self.type}) id={self.req_id}"


from bayserver_core.bay_log import BayLog
from bayserver_core.agent.next_socket_action import NextSocketAction
from bayserver_core.protocol.packet_unpacker import PacketUnPacker
from bayserver_core.util.simple_buffer import SimpleBuffer

from bayserver_docker_fcgi.fcg_packet import FcgPacket

class FcgPacketUnPacker(PacketUnPacker):
    STATE_READ_PREAMBLE = 1  # #state for reading first 8 bytes (from version to reserved)
    STATE_READ_CONTENT = 2  # state for reading content data
    STATE_READ_PADDING = 3  # state for reading padding data
    STATE_END = 4  # End

    def __init__(self, pkt_store, cmd_unpacker):
        self.cmd_unpacker = cmd_unpacker
        self.pkt_store = pkt_store
        self.header_buf = SimpleBuffer()
        self.data_buf = SimpleBuffer()
        self.state = None
        self.version = None
        self.type = None
        self.req_id = None
        self.length = None
        self.padding = None
        self.padding_read_bytes = None
        self.cont_len = None
        self.read_bytes = None
        self.reset()

    def reset(self):
        self.state = FcgPacketUnPacker.STATE_READ_PREAMBLE
        self.version = 0
        self.type = None
        self.req_id = 0
        self.length = 0
        self.padding = 0
        self.padding_read_bytes = 0
        self.cont_len = 0
        self.read_bytes = 0
        self.header_buf.reset()
        self.data_buf.reset()

    def bytes_received(self, buf):
        next_suspend = False
        next_write = False
        pos = 0

        while pos < len(buf):
            while self.state != FcgPacketUnPacker.STATE_END and pos < len(buf):

                if self.state == FcgPacketUnPacker.STATE_READ_PREAMBLE:
                    # preamble read mode
                    length = FcgPacket.PREAMBLE_SIZE - len(self.header_buf)
                    if len(buf) - pos < length:
                        length = len(buf) - pos

                    self.header_buf.put(buf, pos, length)
                    pos += length

                    if len(self.header_buf) == FcgPacket.PREAMBLE_SIZE:
                        self.header_read_done()
                        if self.length == 0:
                            if self.padding == 0:
                                self.change_state(FcgPacketUnPacker.STATE_END)
                            else:
                                self.change_state(FcgPacketUnPacker.STATE_READ_PADDING)

                        else:
                            self.change_state(FcgPacketUnPacker.STATE_READ_CONTENT)


                elif self.state == FcgPacketUnPacker.STATE_READ_CONTENT:
                    # content read mode
                    length = self.length - len(self.data_buf)
                    if length > len(buf) - pos:
                        length = len(buf) - pos

                    if length > 0:
                        self.data_buf.put(buf, pos, length)
                        pos += length

                        if len(self.data_buf) == self.length:
                            if self.padding == 0:
                                self.change_state(FcgPacketUnPacker.STATE_END)
                            else:
                                self.change_state(FcgPacketUnPacker.STATE_READ_PADDING)


                elif self.state == FcgPacketUnPacker.STATE_READ_PADDING:
                    # padding read mode
                    length = self.padding - self.padding_read_bytes

                    if length > len(buf) - pos:
                        length = len(buf) - pos

                    # self.data_buf.put(buf, pos, len)
                    pos += length

                    if length > 0:
                        self.padding_read_bytes += length

                        if self.padding_read_bytes == self.padding:
                            self.change_state(FcgPacketUnPacker.STATE_END)

                else:
                  raise RuntimeError("IllegalState")

            if self.state == FcgPacketUnPacker.STATE_END:
                pkt = self.pkt_store.rent(self.type)
                pkt.req_id = self.req_id
                pkt.new_header_accessor().put_bytes(self.header_buf.buf, 0, len(self.header_buf))
                pkt.new_data_accessor().put_bytes(self.data_buf.buf, 0, len(self.data_buf))

                try:
                    state = self.cmd_unpacker.packet_received(pkt)
                finally:
                    self.pkt_store.Return(pkt)

                self.reset()

                if self.state == NextSocketAction.SUSPEND:
                    next_suspend = True
                elif self.state == NextSocketAction.CONTINUE:
                    pass
                elif self.state == NextSocketAction.WRITE:
                    next_write = True
                elif self.state == NextSocketAction.CLOSE:
                    return state

        if next_write:
            return NextSocketAction.WRITE
        elif next_suspend:
            return NextSocketAction.SUSPEND
        else:
            return NextSocketAction.CONTINUE

    def change_state(self, new_state):
        self.state = new_state

    def header_read_done(self):
        pre = self.header_buf.buf
        self.version = self.byte_to_int(pre[0])
        self.type = self.byte_to_int(pre[1])
        self.req_id = self.bytes_to_int(pre[2], pre[3])
        self.length = self.bytes_to_int(pre[4], pre[5])
        self.padding = self.byte_to_int(pre[6])
        reserved = self.byte_to_int(pre[7])
        BayLog.debug("fcg: read packet header: version=%s type=%d reqId=%d length=%d padding=%d",
                     self.version, self.type, self.req_id, self.length, self.padding)

    def byte_to_int(self, b):
        return b & 0xff

    def bytes_to_int(self, b1, b2):
        return self.byte_to_int(b1) << 8 | self.byte_to_int(b2)



from bayserver_core.util.char_util import CharUtil

from bayserver_docker_fcgi.fcg_command import FcgCommand
from bayserver_docker_fcgi.fcg_type import FcgType


#
#  FCGI spec
#    http://www.mit.edu/~yandros/doc/specs/fcgi-spec.html
#
#  Begin request command format
#          typedef struct {
#              unsigned char roleB1;
#              unsigned char roleB0;
#              unsigned char flags;
#              unsigned char reserved[5];
#          } FCGI_BeginRequestBody;
#
class CmdBeginRequest(FcgCommand):
    FCGI_KEEP_CONN = 1
    FCGI_RESPONDER = 1
    FCGI_AUTHORIZER = 2
    FCGI_FILTER = 3

    RESERVED = [CharUtil.SPACE_BYTE] * 5

    def __init__(self, req_id):
        super().__init__(FcgType.BEGIN_REQUEST, req_id)
        self.role = None
        self.keep_conn = None

    def unpack(self, pkt):
        super().unpack(pkt)

        acc = pkt.new_data_accessor()
        self.role = acc.get_short()
        flags = acc.get_byte()
        self.keep_conn = (flags & CmdBeginRequest.FCGI_KEEP_CONN) != 0

    def pack(self, pkt):
        acc = pkt.new_data_accessor()
        acc.put_short(self.role)
        acc.put_byte(1 if self.keep_conn else 0)
        acc.put_bytes(CmdBeginRequest.RESERVED)

        # must be called from last line
        super().pack(pkt)

    def handle(self, cmd_handler):
        return cmd_handler.handle_begin_request(self)









from bayserver_core.protocol.command import Command

class FcgCommand(Command):

    def __init__(self, typ, req_id):
        super().__init__(typ)
        self.req_id = req_id

    def unpack(self, pkt):
        self.req_id = pkt.req_id

    #
    # Super class method must be called from last line of override method
    # since header cannot be packed before data is constructed
    #
    def pack(self, pkt):
        pkt.req_id = self.req_id
        self.pack_header(pkt)

    def pack_header(self, pkt):
        acc = pkt.new_header_accessor()
        acc.put_byte(pkt.version)
        acc.put_byte(pkt.type)
        acc.put_short(pkt.req_id)
        acc.put_short(pkt.data_len())
        acc.put_byte(0)  # paddinglen
        acc.put_byte(0)  # reserved





import socket

from bayserver_core.bayserver import BayServer
from bayserver_core.bay_log import BayLog
from bayserver_core.http_exception import HttpException
from bayserver_core.bay_message import BayMessage
from bayserver_core.symbol import Symbol

from bayserver_core.agent.next_socket_action import NextSocketAction
from bayserver_core.protocol.protocol_handler_factory import ProtocolHandlerFactory
from bayserver_core.protocol.protocol_exception import ProtocolException
from bayserver_core.tour.tour import Tour
from bayserver_core.tour.tour_req import TourReq
from bayserver_core.tour.req_content_handler import ReqContentHandler
from bayserver_core.docker.base.inbound_handler import InboundHandler

from bayserver_core.util.simple_buffer import SimpleBuffer
from bayserver_core.util.http_status import HttpStatus
from bayserver_core.util.headers import Headers
from bayserver_core.util.http_util import HttpUtil
from bayserver_core.util.cgi_util import CgiUtil
from bayserver_core.util.class_util import ClassUtil

from bayserver_docker_fcgi.fcg_packet import FcgPacket
from bayserver_docker_fcgi.fcg_protocol_handler import FcgProtocolHandler
from bayserver_docker_fcgi.command.cmd_stdout import CmdStdOut
from bayserver_docker_fcgi.command.cmd_end_request import CmdEndRequest


class FcgInboundHandler(FcgProtocolHandler, InboundHandler):

    class InboundProtocolHandlerFactory(ProtocolHandlerFactory):

        def create_protocol_handler(self, pkt_store):
            return FcgInboundHandler(pkt_store)

    STATE_BEGIN_REQUEST = 1
    STATE_READ_PARAMS = 2
    STATE_READ_STDIN = 3

    def __init__(self, pkt_store):
        super().__init__(pkt_store, True)
        self.state = None
        self.cont_len = None
        self.req_id = None
        self.req_keep_alive = None
        self.env = {}
        self.reset_state()

    def __str__(self):
        return ClassUtil.get_local_name(self.__class__)


    ######################################################
    # implements Reusable
    ######################################################
    def reset(self):
        super().reset()
        self.env.clear()
        self.reset_state()

    ######################################################
    # implements InboundHandler
    ######################################################

    def send_res_headers(self, tur):
        BayLog.debug("%s PH:sendHeaders: tur=%s", self.ship, tur)

        scode = tur.res.headers.status
        status = f"{scode} {HttpStatus.description(scode)}"
        tur.res.headers.set(Headers.STATUS, status)

        if BayServer.harbor.trace_header:
            BayLog.info("%s resStatus:%d", tur, tur.res.headers.status)
            for name in tur.res.headers.names():
                for value in tur.res.headers.values(name):
                    BayLog.info("%s resHeader:%s=%s", tur, name, value)

        buf = SimpleBuffer()
        HttpUtil.send_mime_headers(tur.res.headers, buf)
        HttpUtil.send_new_line(buf)
        cmd = CmdStdOut(tur.req.key, buf.buf, 0, buf.length)
        self.command_packer.post(self.ship, cmd)

    def send_res_content(self, tur, bytes, ofs, length, callback):

        cmd = CmdStdOut(tur.req.key, bytes, ofs, length)
        self.command_packer.post(self.ship, cmd, callback)


    def send_end_tour(self, tur, keep_alive, cb):
        BayLog.debug("%s PH:endTour: tur=%s keep=%s", self.ship, tur, keep_alive)

        # Send empty stdout command
        cmd = CmdStdOut(tur.req.key)
        self.command_packer.post(self.ship, cmd)

        # Send end request command
        cmd = CmdEndRequest(tur.req.key)

        def ensure_func():
            if not keep_alive:
                self.command_packer.end(self.ship)

        def callback_func():
            BayLog.debug("%s call back in sendEndTour: tur=%s keep=%s", self, tur, keep_alive)
            ensure_func()
            cb()

        try:
            self.command_packer.post(self.ship, cmd, callback_func)
        except IOError as e:
            BayLog.debug("%s post failed in sendEndTour: tur=%s keep=%s", self, tur, keep_alive)
            ensure_func()
            raise e

    def send_req_protocol_error(self, err):
        tur = self.ship.get_error_tour()
        tur.res.send_error(Tour.TOUR_ID_NOCHECK, HttpStatus.BAD_REQUEST, err.message, err)
        return True


    ######################################################
    # implements FcgCommandHandler
    ######################################################
    def handle_begin_request(self, cmd):
        BayLog.debug("%s handle_begin_request req_id=%d} keep=%s", self.ship, cmd.req_id, cmd.keep_conn)

        if self.state != FcgInboundHandler.STATE_BEGIN_REQUEST:
            raise ProtocolException("Invalid FCGI command: %d state=%d", cmd.type, self.state)

        self.check_req_id(cmd.req_id)

        self.req_id = cmd.req_id
        BayLog.debug("%s begin_req get_tour req_id=%d", self.ship, cmd.req_id)
        tur = self.ship.get_tour(cmd.req_id)
        if tur is None:
            BayLog.error(BayMessage.get(Symbol.INT_NO_MORE_TOURS))
            tur = self.ship.get_tour(cmd.req_id, True)
            tur.res.send_error(Tour.TOUR_ID_NOCHECK, HttpStatus.SERVICE_UNAVAILABLE, "No available tours")
            return NextSocketAction.CONTINUE

        self.req_keep_alive = cmd.keep_conn
        self.change_state(FcgInboundHandler.STATE_READ_PARAMS)
        return NextSocketAction.CONTINUE

    def handle_end_request(self, cmd):
        raise ProtocolException("Invalid FCGI command: %d", cmd.type)

    def handle_params(self, cmd):
        BayLog.debug("%s handle_params req_id=%d", self.ship, cmd.req_id)

        if self.state != FcgInboundHandler.STATE_READ_PARAMS:
            raise ProtocolException("Invalid FCGI command: %d state=%d", cmd.type, self.state)

        self.check_req_id(cmd.req_id)

        BayLog.debug("%s handle_param get_tour req_id=%d", self.ship, cmd.req_id)
        tur = self.ship.get_tour(cmd.req_id)

        if len(cmd.params) == 0:
            # Header completed

            # check keep-alive
            #  keep-alive flag of BeginRequest has high priority
            if self.req_keep_alive:
                if not tur.req.headers.contains(Headers.CONNECTION):
                    tur.req.headers.set(Headers.CONNECTION, "Keep-Alive")
                else:
                    tur.req.headers.set(Headers.CONNECTION, "Close")

            req_cont_len = tur.req.headers.content_length()

            BayLog.debug("%s read header method=%s protocol=%s uri=%s contlen=%d",
                         self.ship, tur.req.method, tur.req.protocol, tur.req.uri, self.cont_len)

            if BayServer.harbor.trace_header:
                for nv in cmd.params:
                    BayLog.info("%s  reqHeader: %s=%s", tur, nv[0], nv[1])

            if req_cont_len > 0:
                sid = self.ship.ship_id
                def callback(length, resume):
                    if resume:
                        self.ship.resume(sid)

                tur.req.set_consume_listener(req_cont_len, callback)

            self.change_state(FcgInboundHandler.STATE_READ_STDIN)
            try:
                self.start_tour(tur)
            except HttpException as e:
                BayLog.debug("%s Http error occurred: %s", self.ship, e)

                if req_cont_len <= 0:
                    # no post data
                    tur.res.send_http_exception(Tour.TOUR_ID_NOCHECK, e)
                    self.change_state(FcgInboundHandler.STATE_READ_STDIN)
                    return NextSocketAction.CONTINUE
                else:
                    # Delay send
                    self.change_state(FcgInboundHandler.STATE_READ_STDIN)
                    tur.error = e
                    tur.req.set_content_handler(ReqContentHandler.dev_null)
                    return NextSocketAction.CONTINUE

        else:
            if BayServer.harbor.trace_header:
                BayLog.info("%s Read FcgiParam", tur)

            for nv in cmd.params:
                name = nv[0]
                value = nv[1]
                if BayServer.harbor.trace_header:
                    BayLog.info("%s  param: %s=%s", tur, name, value);
                self.env[name] = value

                if name.startswith("HTTP_"):
                    hname = name[5:]
                    tur.req.headers.add(hname, value)
                elif name == "CONTENT_TYPE":
                    tur.req.headers.add(Headers.CONTENT_TYPE, value)
                elif name == "CONTENT_LENGTH":
                    tur.req.headers.add(Headers.CONTENT_LENGTH, value)
                elif name == "HTTPS":
                    tur.is_secure = value.lower() == "on"

            tur.req.uri = self.env["REQUEST_URI"]
            tur.req.protocol = self.env["SERVER_PROTOCOL"]
            tur.req.method = self.env["REQUEST_METHOD"]

            return NextSocketAction.CONTINUE

    def handle_stderr(self, cmd):
        raise ProtocolException("Invalid FCGI command: %d", cmd.type)

    def handle_stdin(self, cmd):
        BayLog.debug("%s handle_stdin req_id=%d len=%d", self.ship, cmd.req_id, cmd.length)

        if self.state != FcgInboundHandler.STATE_READ_STDIN:
            raise ProtocolException("Invalid FCGI command: %d state=%d", cmd.type, self.state)


        self.check_req_id(cmd.req_id)

        tur = self.ship.get_tour(cmd.req_id)
        if cmd.length == 0:
            #  request content completed
            if tur.error:
                # Error has occurred on header completed
                tur.res.send_http_exception(Tour.TOUR_ID_NOCHECK, tur.error)
                self.reset_state()
                return NextSocketAction.WRITE
            else:
                try:
                    self.end_req_content(Tour.TOUR_ID_NOCHECK, tur)
                    return NextSocketAction.CONTINUE
                except HttpException as e:
                    tur.res.send_http_exception(Tour.TOUR_ID_NOCHECK, e)
                    return NextSocketAction.WRITE

        else:
            success = tur.req.post_content(Tour.TOUR_ID_NOCHECK, cmd.data, cmd.start, cmd.length)

            if not success:
                return NextSocketAction.SUSPEND
            else:
                return NextSocketAction.CONTINUE

    def handle_stdout(self, cmd):
        raise ProtocolException("Invalid FCGI command: %d", cmd.type)

    #
    # private
    #
    def check_req_id(self, received_id):
        if received_id == FcgPacket.FCGI_NULL_REQUEST_ID:
            raise ProtocolException("Invalid request id: %d", received_id)

        if self.req_id == FcgPacket.FCGI_NULL_REQUEST_ID:
            self.req_id = received_id

        if self.req_id != received_id:
            BayLog.error("%s invalid request id: received=%d reqId=%d", self.ship(), received_id, self.req_id)
            raise ProtocolException("Invalid request id: %d", received_id)

    def change_state(self, new_state):
        self.state = new_state

    def reset_state(self):
        self.change_state(FcgInboundHandler.STATE_BEGIN_REQUEST)
        self.req_id = FcgPacket.FCGI_NULL_REQUEST_ID
        self.cont_len = 0

    def end_req_content(self, check_id, tur):
        tur.req.end_content(check_id)
        self.reset_state()

    def start_tour(self, tur):
        HttpUtil.parse_host_port(tur, 443 if tur.is_secure else 80)
        HttpUtil.parse_authorization(tur)

        tur.req.remote_port = int(self.env[CgiUtil.REMOTE_PORT])
        tur.req.remote_address = self.env[CgiUtil.REMOTE_ADDR]
        tur.req.remote_host_func = lambda: HttpUtil.resolve_remote_host(tur.req.remote_address)

        tur.req.server_name = self.env[CgiUtil.SERVER_NAME]
        tur.req.server_address = self.env[CgiUtil.SERVER_ADDR]
        tur.req.server_port = int(self.env[CgiUtil.SERVER_PORT])

        tur.go()




from setuptools import setup, find_packages

print("packages: " + str(find_packages()))

setup(
    name='bayserver-docker-fcgi',
    version='2.3.4',
    packages=find_packages(),
    package_data={
        '': ['LICENSE.BAYKIT', 'README.md'],
    },  
    install_requires=[
      "bayserver-core==2.3.4",
    ],
    author='Michisuke-P',
    author_email='michisukep@gmail.com',
    description='FCGI Docker for BayServer',
    license='MIT',
    python_requires=">=3.7",
    url='https://baykit.yokohama/',
)




class FcgParams:
    PATH = "PATH"
    CONTEXT_DOCUMENT_ROOT = "CONTEXT_DOCUMENT_ROOT"
    CONTEXT_PREFIX = "CONTEXT_PREFIX"
    CONTENT_TYPE = "CONTENT_TYPE"
    CONTENT_LENGTH = "CONTENT_LENGTH"
    DOCUMENT_ROOT = "DOCUMENT_ROOT"
    GATEWAY_INTERFACE = "GATEWAY_INTERFACE"
    HTTP_ACCEPT = "HTTP_ACCEPT"
    HTTP_ACCEPT_ENCODING = "HTTP_ACCEPT_ENCODING"
    HTTP_ACCEPT_LANGUAGE = "HTTP_ACCEPT_LANGUAGE"
    HTTP_CACHE_CONTROL = "HTTP_CACHE_CONTROL"
    HTTP_COOKIE = "HTTP_COOKIE"
    HTTP_CONNECTION = "HTTP_CONNECTION"
    HTTP_HOST = "HTTP_HOST"
    HTTP_USER_AGENT = "HTTP_USER_AGENT"
    HTTP_UPGRADE_INSECURE_REQUESTS = "HTTP_UPGRADE_INSECURE_REQUESTS"
    QUERY_STRING = "QUERY_STRING"
    REMOTE_ADDR = "REMOTE_ADDR"
    REQUEST_SCHEME = "REQUEST_SCHEME"
    REQUEST_URI = "REQUEST_URI"
    REMOTE_PORT = "REMOTE_PORT"
    REQUEST_METHOD = "REQUEST_METHOD"
    SCRIPT_NAME = "SCRIPT_NAME"
    SCRIPT_FILENAME = "SCRIPT_FILENAME"
    SERVER_ADDR = "SERVER_ADDR"
    SERVER_ADMIN = "SERVER_ADMIN"
    SERVER_NAME = "SERVER_NAME"
    SERVER_PORT = "SERVER_PORT"
    SERVER_PROTOCOL = "SERVER_PROTOCOL"
    SERVER_SIGNATURE = "SERVER_SIGNATURE"
    SERVER_SOFTWARE = "SERVER_SOFTWARE"
    UNIQUE_ID = "UNIQUE_ID"
    X_FORWARDED_FOR = "X-Forwarded-For"
    X_FORWARDED_PROTO = "X-Forwarded-Proto"
    X_FORWARDED_PORT = "X-Forwarded-Port"



from bayserver_core.bay_log import BayLog

from bayserver_core.agent.transporter.plain_transporter import PlainTransporter
from bayserver_core.protocol.packet_store import PacketStore
from bayserver_core.protocol.protocol_handler_store import ProtocolHandlerStore
from bayserver_core.util.io_util import IOUtil

from bayserver_core.docker.warp.warp_docker import WarpDocker

from bayserver_docker_fcgi.fcg_docker import FcgDocker
from bayserver_docker_fcgi.fcg_packet_factory import FcgPacketFactory
from bayserver_docker_fcgi.fcg_warp_handler import FcgWarpHandler

class FcgWarpDocker(WarpDocker, FcgDocker):

    def __init__(self):
        super().__init__()
        self.script_base = None
        self.doc_root = None


    ######################################################
    # Implements Docker
    ######################################################
    def init(self, elm, parent):
        super().init(elm, parent)

        if self.script_base is None:
            BayLog.warn("FCGI: docRoot is not specified")

    ######################################################
    # Implements DockerBase
    ######################################################

    def init_key_val(self, kv):
        key = kv.key.lower()
        if key == "scritbase":
            self.script_base = kv.value
        elif key == "docroot":
            self.doc_root = kv.value
        else:
            return super().init_key_val(kv)

        return True

    ######################################################
    # Implements WarpDocker
    ######################################################
    def secure(self):
        return False

    ######################################################
    # Implements WarpDockerBase
    ######################################################
    def protocol(self):
        return FcgDocker.PROTO_NAME

    def new_transporter(self, agt, skt):
        return PlainTransporter(False, IOUtil.get_sock_recv_buf_size(skt))

    ######################################################
    # Class initializer
    ######################################################
    PacketStore.register_protocol(
        FcgDocker.PROTO_NAME,
        FcgPacketFactory())
    ProtocolHandlerStore.register_protocol(
        FcgDocker.PROTO_NAME,
        False,
        FcgWarpHandler.WarpProtocolHandlerFactory())
